package lsedit;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;

import java.util.*;
import java.io.*; 
import javax.swing.*;

// EntityInstance class
// 
// Provides most of the data and methods for an entity.
// Can't be instantiated directly. Requires a subclass to provide
// the draw and other functions based on a shape (e.g. box).
//

public abstract class EntityInstance extends LandscapeObject /* extends JComponent */ implements Icon, DiagramCoordinates {

	// Final values
	
	public final static int SMALL_FONT = 0;
	public final static int REG_FONT   = 1;

	public final static String LABEL_ID =			 "label";
	public final static String XPOSITION_ID =		 "x";
	public final static String YPOSITION_ID =		 "y";
	public final static String WIDTH_ID =			 "width";
	public final static String HEIGHT_ID =			 "height";

	public final static String XRELPOSITION_ID =	 "xrel";
	public final static String YRELPOSITION_ID =	 "yrel";
	public final static String WIDTHREL_ID =		 "widthrel";
	public final static String HEIGHTREL_ID =		 "heightrel";

	public final static String IN_ELISION_ID =		 "elision";
	public final static String OUT_ELISION_ID =		 "outelision";
	public final static String CLIENT_ELISION_ID =	 "clientelision";
	public final static String SUPPLIER_ELISION_ID = "supplierelision";
	public final static String INTERNAL_ELISION_ID = "internalelision";
	public final static String AGG_ELISION_ID =		 "aggelision";
	public final static String LINK_ID =			 "navlink";
	public final static String INPOINT_ID =			 "inpoints";
	public final static String OUTPOINT_ID =		 "outpoints";
	public final static String LEFTPOINT_ID =		 "leftpoints";
	public final static String RIGHTPOINT_ID =		 "rightpoints";
	public final static String DESC_ID =			 "description";
	public final static String TITLE_ID =			 "title";
	public final static String FONTDELTA_ID =		 "fontdelta";

	protected final static int RSZ_NONE = -1;
	protected final static int RSZ_NW	= 0;
	protected final static int RSZ_N	= 1;
	protected final static int RSZ_NE	= 2; 
	protected final static int RSZ_E	= 3;
	protected final static int RSZ_SE	= 4;
	protected final static int RSZ_S	= 5;
	protected final static int RSZ_SW	= 6;
	protected final static int RSZ_W	= 7;

	protected final static int DEF_CONT_CAP  = 35; // Default contains capacity
	protected final static int ROOT_CONT_CAP = 1000; // root contains capacity

	// Annotation tab constants

	protected final static double TAB_HEIGHT = 16.0; 

	protected final static int MOUSE_NEAR_EDGE_THRESHOLD = 4;
	protected final static int SEP_THRESHOLD = 8;

	// Static font info 

	protected final static int MARGIN = 5; 
	protected final static int MIN_HEIGHT = 10;
	protected final static int MIN_WIDTH = 30;

	protected final static int CONTENTS_FLAG_DIM = 8;

	protected final static String OPEN_FAMILY = "Helvetica";
	protected final static int	  OPEN_FONT_SIZE = 12;
	protected final static Font openFont = new Font(OPEN_FAMILY, Font.PLAIN, OPEN_FONT_SIZE);

	protected final static String CLOSED_FAMILY = "Helvetica";
	protected final static int	  CLOSED_FONT_MIN_SIZE = 6; 
	protected final static int	  CLOSED_FONT_SMALL_SIZE = 10; 
	protected final static int	  NUM_FONTS = 25;
	protected final static int	  START_FONT_NUM = 6;
	public    final static Font smallFont =	new Font(CLOSED_FAMILY, Font.PLAIN, CLOSED_FONT_SMALL_SIZE);

	// Bits set in m_mark
	
	public final static int	DIAGRAM_MARK		= 0x01;		// This entity is visible in the main diagram  (at present)
	public final static int	CLIENT_MARK			= 0x02;		// This entity is in the set of clients   (at present)
	public final static int SUPPLIER_MARK		= 0x04;		// This entity is in the set of suppliers (at present)

	public final static int HIGHLIGHT_EDGE_MARK	= 0x01000;		// Highlight the edge of the object
	public final static int DRAW_CLIENT_MARK	= 0x02000;
	public final static int DRAW_SUPPLIER_MARK	= 0x04000;
	public final static int REDBOX_MARK         = 0x08000;	// Mark this box in deep red for (forward/back) edge tracing     
	public final static int	GROUP_MARK			= 0x10000;	// This entity is part of a group
	public final static int GROUPKEY_MARK		= 0x20000;	// This entity is the key entity in the group
	public final static int	OPEN_MARK			= 0x40000;	// This entity is open
	public final static int CLOSED_MARK			= 0x80000;	// This entity is closed

	public final static int IN_DIAGRAM          = CLIENT_MARK|SUPPLIER_MARK|DIAGRAM_MARK;
	public final static int PRESENTATION_MARKS  = HIGHLIGHT_EDGE_MARK | REDBOX_MARK | GROUP_MARK | GROUPKEY_MARK;

	public final static int	IN_TREE_MARK		= 0x100000;
	public final static int	LIFTED_MARK			= 0x200000;
	
	private int m_mark = 0;

	protected EntityClass m_parentClass;				// The class this entity belongs to

	protected static Font closedFont[]    = new Font[NUM_FONTS];


	protected static boolean mungeIdToLabel = true;

	// First order attributes

	protected Vector	 m_dstElision      = new Vector();
	protected Vector	 m_srcElision      = new Vector();
	protected Vector	 m_enteringElision = new Vector();
	protected Vector	 m_exitingElision  = new Vector();
	protected Vector	 m_internalElision = new Vector();

	protected EdgePoint[] topPoints, bottomPoints, leftPoints, rightPoints;

	// Changed width and height to have some [reasonable?] default IJD

	protected int		m_diagramX         = Integer.MIN_VALUE;
	protected int		m_diagramY         = Integer.MIN_VALUE;

	// These are the coordinates with respect to the parent node

	private double		m_xrelLocal	     = 1.0/32.0;
	private double		m_yrelLocal      = 1.0/32.0;
	private double		m_widthrelLocal  = 15.0/16.0;
	private	double		m_heightrelLocal = 15.0/16.0;

	// These are the top left coordinates of the object on the screen

	protected int		 m_fontDelta     = 0;

	protected int		 m_preorder;
	protected int		 m_postorder;
	
	/* keep tight control over the group flags so can draw changes at a low level */

	// Relation info

	// Our container (null -> $ROOT)

	protected EntityInstance m_containedBy;

	// Entities contained by this one

	protected Hashtable containsList;			

	protected Vector m_srcRelList    = new Vector();	// List of relations for which this entity is the source 
	protected Vector m_dstRelList    = new Vector();	// List of relations for which this entity is the destination 
	protected Vector m_srcLiftedList = new Vector();	// List of relations whose source is lifted to leave this entity
	protected Vector m_dstLiftedList = new Vector();	// List of relations whose destination is lifted to address this entity

	// Temp Rectangle storage

	Rectangle tempLyt;

	// Stat gathering

	protected Object	m_ancilliary;

/*
	private int m_debug_x      = Integer.MIN_VALUE;
	private int m_debug_y      = Integer.MIN_VALUE;
	private int m_debug_width  = Integer.MIN_VALUE;
	private int m_debug_height = Integer.MIN_VALUE;
*/
	 
	// -----------------
	// Protected methods
	// -----------------

	protected void setFont(Graphics g, int type) 
	{
		switch(type) {
		case SMALL_FONT:
			g.setFont(smallFont);
			return;
		default:
			g.setFont(closedFont[START_FONT_NUM + m_fontDelta]);
			return;
		}
	}

	protected void setFontDelta(int fd) 
	{
		// Keep it in bounds
//		System.out.print("Changing font from size " + m_fontDelta);
		m_fontDelta = Math.max(-START_FONT_NUM, Math.min(fd, NUM_FONTS-START_FONT_NUM-1));
//		System.out.println(" to " + m_fontDelta);
		repaint();
	}

	protected double tabHeight() 
	{
		return getHeight()/3-2;
	}

	protected void setGraphicsColor(Graphics g, Color explicit) 
	{
		if (explicit == null) {
			g.setColor(getObjectColor());
		} else {
			g.setColor(explicit);
	}	}

	public LandscapeObject derivedFrom(int i)
	{
		return((i == 0) ? m_parentClass : null);
	}

	public Color getLabelColour() 
	{
		if (isMarked(REDBOX_MARK) && !isOpen()) {
			return Color.yellow;
		}
		return getLabelColor();
	}

	//
	// Parse an attribute record to set the appropriate edge point factors
	// supplied in an attribute record.

	protected void parsePoints(EdgePoint[] ept, Attribute attr) 
	{
		AttributeValueItem avi = attr.avi;

		while (avi != null) {
			RelationClass rc = dg.nameToRelationClass(avi.value);

			if (rc != null) {
				int nid = rc.getNid();

				double wf = Util.parseReal(avi.next.value);
				double hf = Util.parseReal(avi.next.next.value);

				if (ept[nid] == null) {
					ept[nid] = new EdgePoint(this, rc, wf, hf);
				} else {
					if (rc != ept[nid].getRc()) {
						MsgOut.println("Class nid mismatch");
					}
					ept[nid].m_wf = wf;
					ept[nid].m_hf = hf;
				}

				ept[nid].isDefault = false;
				avi = avi.nextList;
			}
		}
	}

	static protected Hashtable est = new Hashtable();

	protected void processElision(Attribute attr, Vector list) 
	{
		AttributeValueItem av = attr.avi;

		while(av != null) {
			String str = (String) est.get(av.value);
			if (str == null) {
				str = av.value;
				est.put(str, str);
			}
			if (!list.contains(str)) {
				list.addElement(str);
			}
			av = av.next;
		}
	}

	protected void processRawElision(String[] sl, Vector list) 
	{
		for (int i = 0; i<sl.length; i++) {
			list.addElement(sl[i]);
		}
	}

	//
	// Parse the values of any first order attributes into their storage.
	// 

	// First thing called by EntityInstance::AddAttributes
	// If the attribute is a known member of storage -- don't save attribute -- just store value

	protected boolean processFirstOrderAttributes(Attribute attr)
	{
		boolean hasVal = (attr.avi != null);

		if (attr.hasId(XPOSITION_ID)) {			
			if (hasVal) {
				setxLocal(attr.parseReal());		// Legacy code
				attr.id = XRELPOSITION_ID;			// So right name gets registered
			}
			return true; 
		}

		if (attr.hasId(YPOSITION_ID)) {
			if (hasVal) {
				setyLocal(attr.parseReal());		// Legacy code
				attr.id = YRELPOSITION_ID;
			}
			return true; 
		}

		if (attr.hasId(WIDTH_ID)) {
			if (hasVal) {
				setwidthLocal(attr.parseReal());	// Legacy code
				attr.id = WIDTHREL_ID;
			}
			return true; 
		}

		if (attr.hasId(HEIGHT_ID)) {
			if (hasVal) {
				setheightLocal(attr.parseReal());	// Legacy code
				attr.id = HEIGHTREL_ID;
			}
			return true; 
		}

		if (attr.hasId(XRELPOSITION_ID)) {
			if (hasVal) {
				setxRelLocal(attr.parseReal());
			}
			return true; 
		}

		if (attr.hasId(YRELPOSITION_ID)) {
			if (hasVal) {
				setyRelLocal(attr.parseReal());
			}
			return true; 
		}

		if (attr.hasId(WIDTHREL_ID)) {
			if (hasVal) {
				setwidthRelLocal(attr.parseReal());
			}
			return true; 
		}

		if (attr.hasId(HEIGHTREL_ID)) {
			if (hasVal) {
				setheightRelLocal(attr.parseReal());
			}
			return true; 
		}

		if (attr.hasId(LABEL_ID)) {
			if (hasVal) {
				m_label = attr.parseString();
			}
			return true; 
		}

		if (attr.hasId(DESC_ID)) {
			if (hasVal) { 
				description = attr.parseString();
			}
			return true; 
		} 

		if (attr.hasId(TITLE_ID)) {
			if (hasVal) {
				title = attr.parseString();
			}
			return true;
		}

		if (attr.hasId(FONTDELTA_ID)) {
			if (hasVal) {
				setFontDelta(attr.parseInt());
			}
			return true; 
		}

		if (attr.hasId(COLOUR_ID)) {
			if (hasVal) {
				setObjectColor(attr.parseColour());
			}
			return true;
		}

		if (attr.hasId(OPEN_COLOUR_ID)) {
			if (hasVal) {
				setObjectColorWhenOpen(attr.parseColour());
			}
			return true;
		}

		if (attr.hasId(LABEL_COLOUR_ID)) {
			if (hasVal) {
				setLabelColor(attr.parseColour());
			}
			return true;
		}

		if (attr.hasId(IN_ELISION_ID)) {
			if (hasVal) { 
				processElision(attr, m_dstElision); 
			}
			return true; 
		}

		if (attr.hasId(OUT_ELISION_ID)) {
			if (hasVal) {
				processElision(attr, m_srcElision);
			}
			return true;
		}

		if (attr.hasId(CLIENT_ELISION_ID)) {
			if (hasVal) {
				processElision(attr, m_enteringElision);
			}
			return true;
		}

		if (attr.hasId(SUPPLIER_ELISION_ID)) {
			if (hasVal) {
				processElision(attr, m_exitingElision);
			}
			return true;
		}

		if (attr.hasId(INTERNAL_ELISION_ID)) {
			if (hasVal) {
				processElision(attr, m_internalElision);
			}
			return true;
		}

		if (attr.hasId(INPOINT_ID)) {
			if (hasVal) {
				if (topPoints == null) {
					topPoints = new EdgePoint[dg.numRelationClasses];
				}
				parsePoints(topPoints, attr);
			}
			return true; 
		}

		if (attr.hasId(OUTPOINT_ID)) {
			if (hasVal) {
				if (bottomPoints == null) {
					bottomPoints = new EdgePoint[dg.numRelationClasses];
				}
				parsePoints(bottomPoints, attr); 
			}
			return true; 
		}

		if (attr.hasId(LEFTPOINT_ID)) {
			if (hasVal) {
				if (leftPoints == null) {
					leftPoints = new EdgePoint[dg.numRelationClasses];
				}
				parsePoints(leftPoints, attr); 
			}
			return true; 
		}

		if (attr.hasId(RIGHTPOINT_ID)) {
			if (hasVal) {
				if (rightPoints == null) {
					rightPoints = new EdgePoint[dg.numRelationClasses];
				}
				parsePoints(rightPoints, attr); 
			}
			return true; 
		}

		return false; 
	}

	protected void doEntityDelete() 
	{
		// Remove m_dstRelList relations from entities on our m_srcRelList.
		// Remove m_srcRelList relations from entities on our m_dstRelList. 
		// Remove entity from parent's contain list. 

		Enumeration		en;
		RelationInstance ri;

		for (en = m_srcRelList.elements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();
			ri.getDst().removeDstRelation(ri); 
		} 

		for (en = m_dstRelList.elements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();
			ri.getSrc().removeSrcRelation(ri);
		}
		m_containedBy.removeContainment(this);
	}

	// Determine the bounding box of all children

	public RelRectangle getChildrenRelBoundingBox() 
	{
		Enumeration		en;
		EntityInstance	e;
		double			x1, x2, y1, y2;
		double			temp;
		boolean			seen;

		x1   = Integer.MAX_VALUE;
		x2   = Integer.MIN_VALUE;
		y1   = Integer.MAX_VALUE;
		y2   = Integer.MIN_VALUE;
		seen = false;

		for (en = getChildren(); en.hasMoreElements(); ) {
			seen = true;
			e = (EntityInstance) en.nextElement();

			temp = e.xRelLocal();
			if (temp < x1) {
				x1 = temp;
			}

			temp = e.yRelLocal();
			if (temp < y1) {
				y1 = temp;
			}

			temp = e.xRelLocal() + e.widthRelLocal();
			if (temp > x2) {
				x2 = temp;
			}

			temp = e.yRelLocal() + e.heightRelLocal();
			if (temp > y2) {
				y2 = temp;
		}	}

		if (!seen) {
			MsgOut.dprintln("No children");
			return(null);
		} 
		return new RelRectangle(x1, y1, x2-x1, y2-y1);
	}

	protected String endSet[] = { 
		".ss",
		".dup",
		"__funcdef",
		"__funcdcl",
		"__vardef",
		"__vardcl",
		"__macro", 
		"__struct", 
		"__union", 
		"__enum", 
		"__type"
	}; 

	public String mungeId(String id) 
	{
		if (mungeIdToLabel && id.length() > 0) {

			for (int i = 0; i < endSet.length; i++) {
				if (id.endsWith(endSet[i])) {
					id = id.substring(0, id.length()-endSet[i].length());
					break;
				}
			}

			if (id.charAt(0) != '<') {
				int lpi = id.lastIndexOf('/');

				if (lpi < 0) {
					lpi = id.lastIndexOf('\\');
				}
				if (lpi >= 0 && id.length() != lpi+1) {
					return id.substring(lpi+1);
				}
			}
		}
		return id;
	}



	// --------------

	// Static methods

	// -------------- 


	/* Allocates helvetica fonts of increasing point size */

	static public void generateFonts() {

		for (int i=0; i<NUM_FONTS; i++) {
			closedFont[i] =	new Font(CLOSED_FAMILY, Font.PLAIN, CLOSED_FONT_MIN_SIZE+i /* Point size */);
		}
	}

	static public void setLabelMungeFlag(boolean state) 
	{
		mungeIdToLabel = state;
	}

	// ----------------
	// Abstract methods
	// ----------------

	// Calculate the intercept between the entity and the line between
	// points p1 and p2, where p1 is in the entity.

	abstract public RealPoint calculateIntercept(RealPoint p1, RealPoint p2);

	abstract public boolean isPointOverIO(EdgePoint pt, int x, int y);
	abstract public int		overResizeTab(int x, int y);

	abstract public void drawHighlight();
	abstract public void undrawHighlight();

	abstract public void moveCardinals();

	abstract public EdgePoint getPoint(RelationClass rc, int side);
	abstract public EdgePoint getOutPoint(RelationInstance ri, int edge_mode, Rectangle srcLyt, Rectangle dstLyt);

	abstract public EdgePoint getMouseOverIO(int x, int y);
	abstract public void paintMap(Graphics g, int x, int y, int width, int height, EntityInstance onPath, int depth);
	abstract public void addToDiagram(JComponent under);

	// --------------
	// Public methods
	// --------------

	protected EntityInstance(Diagram dg) 
	{
		this.dg = dg;
	}

	public void setup(EntityClass parentClass, String id, Diagram dg) 
	{
		m_label = mungeId(id); // Default
		this.dg = dg;
		this.id = id;
		setParentClass(parentClass);

		// A default layout
		
		containsList = new Hashtable(id.equals(Diagram.ROOT_ID) ? ROOT_CONT_CAP : DEF_CONT_CAP);
//		setBounds(10, 10, 20, 10);
	}

 

	public EntityInstance(EntityClass parentClass, String id, Diagram dg) 
	{
		setup(parentClass, id, dg);
	}

	// Destroy just the entity, leaving the children 

	public void deleteJustMe() 
	{
		Enumeration en;

		// Promote our children

		for (en = getChildren(); en.hasMoreElements(); ) {
			EntityInstance e = (EntityInstance) en.nextElement();
			m_containedBy.addContainment(e);
		}
		doEntityDelete(); 
		m_containedBy.rescaleChildren();
	}

	// Destroy entity and all it contains 

	public void deleteMe() 
	{
		// deleteAll on our children then ourself

		Enumeration en;

		for (en = getChildren(); en.hasMoreElements(); ) {
			EntityInstance e = (EntityInstance) en.nextElement();
			e.deleteMe();
		} 
		doEntityDelete();
	}

	public void addContainment(EntityInstance e) 
	{
		// This entity contains the passed entity
		// Put on this' contains list

		if (containsList.put(e.id, e) != null) {
			MsgOut.println("Entity " + id + " already contains " + e.id);
		} else {
			e.m_containedBy = this;
			if (e.inDiagram()) {
				nandMark(OPEN_MARK|CLOSED_MARK);
				add(e);		// Add e back into new containment
		}	}
	}

	public boolean removeContainment(EntityInstance e) 
	{
		if (containsList.containsKey(e.id)) {
			containsList.remove(e.id);
			remove(e);		// Remove e from old containment
			return true;
		}
		MsgOut.println("Entity " + id + " doesn't contain " + e.id);
		return false;
	}



	public void moveEntityContainment(EntityInstance newContainer) 
	{
		// Transfer containment iff we need to

		if (newContainer != m_containedBy) {
/*
			System.out.println("moveEntityContainment");
			java.lang.Thread.dumpStack();
			System.out.println("-----");
 */

			m_containedBy.removeContainment(this);
			newContainer.addContainment(this); 
		}
	}



	public void addRelation(RelationInstance ri, EntityInstance dst) 
	{
		// Should check to see if relation is already present
		addSrcRelation(ri); 
		dst.addDestRelation(ri); 
	}

	public RelationInstance getRelation(RelationClass rc, EntityInstance dst) 
	{
		// Return relation instance if present

		Enumeration		en;
		RelationInstance ri;

		for (en = m_srcRelList.elements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();
			if (ri.getRelationClass() == rc && ri.getDst() == dst) {
				return ri;
		}	}
		return null;
	}

	public void addSrcRelation(RelationInstance ri) 
	{
		m_srcRelList.addElement(ri); 
	}

	public void removeSrcRelation(RelationInstance ri) 
	{
		m_srcRelList.removeElement(ri); 
	}

	public void addDestRelation(RelationInstance ri) {

		m_dstRelList.addElement(ri); 

	}

	public void removeDstRelation(RelationInstance ri) 
	{
		m_dstRelList.removeElement(ri);

	}

	public boolean isRelationPresent(RelationClass rc, EntityInstance dst) 
	{
		Enumeration		 en;
		RelationInstance ri;

		for (en = m_srcRelList.elements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();
			if (ri.isSameRelation(rc, this, dst)) {
				return true; 
			}
		}
		return false; 
	}

	public void addAttribute(Attribute attr) {

		// Process first order attributes, and add all others 
		// to an attribute database

		if (processFirstOrderAttributes(attr)) {
			return;
		}
		super.addAttribute(attr);
	}

	public boolean addRawAttribute(String id, int val) {

		if (id.equals(FONTDELTA_ID)) {
			setFontDelta(val);
			return true; 
		}
		return false;
	}

	public boolean addRawAttribute(String id, double val) {

		if (id.equals(XPOSITION_ID)) {			// Legacy code
			setxLocal(val);						// Legacy code
			return true; 
		}
		else if (id.equals(YPOSITION_ID)) {
			setyLocal(val);						// Legacy code
			return true; 
		}
		else if (id.equals(WIDTH_ID)) {
			setwidthLocal(val);					// Legacy code
			return true; 
		}
		else if (id.equals(HEIGHT_ID)) {
			setheightLocal(val);				// Legacy code
			return true; 
		}
		if (id.equals(XRELPOSITION_ID)) {
			setxRelLocal(val);
			return true; 
		}
		else if (id.equals(YRELPOSITION_ID)) {
			setyRelLocal(val);
			return true; 
		}
		else if (id.equals(WIDTHREL_ID)) {
			setwidthRelLocal(val);
			return true; 
		}
		else if (id.equals(HEIGHTREL_ID)) {
			setheightRelLocal(val);
			return true; 
		}
		return false;
	}

	public boolean addRawAttribute(String id, String val) {

		if (id.equals(LABEL_ID)) {
			m_label = val;
			return true; 
		}

		else if (id.equals(DESC_ID)) {
			description = val;
			return true; 
		} 
		else if (id.equals(TITLE_ID)) {
			title = val;
			return true;
		}
		return super.addRawAttribute(id, val);
	}

	public boolean addRawAttribute(String id, int[] il) {
		return false;
	}



	public boolean addRawAttribute(String id, double[] dl) {
		if (id.equals(COLOUR_ID)) {
			setObjectColor( new Color((float) dl[0], (float) dl[1], (float) dl[2]));
			return true;
		}
		if (id.equals(LABEL_COLOUR_ID)) {
			setLabelColor( new Color((float) dl[0], (float) dl[1], (float) dl[2]) );
			return true;
		}
		if (id.equals(OPEN_COLOUR_ID)) {
			setObjectColorWhenOpen( new Color((float) dl[0], (float) dl[1], (float) dl[2]));
			return true;
		}
		return false;
	}



	public boolean addRawAttribute(String id, String[] sl) {

		if (id.equals(IN_ELISION_ID)) {
			processRawElision(sl, m_dstElision); 
			return true; 
		}
		else if (id.equals(OUT_ELISION_ID)) {
			processRawElision(sl, m_srcElision);
			return true;
		}
		else if (id.equals(CLIENT_ELISION_ID)) {
			processRawElision(sl, m_enteringElision);
			return true;
		}
		else if (id.equals(SUPPLIER_ELISION_ID)) {
			processRawElision(sl, m_exitingElision);
			return true;
		}
		else if (id.equals(INTERNAL_ELISION_ID)) {
			processRawElision(sl, m_internalElision);
			return true;
		}

		return super.addRawAttribute(id, sl);
	}

	public void assignAttributes(Attribute attr) {

		// Passed list of attribute assignments

		while (attr != null) {

			// For each attribute assignment, find the variable,
			// and assign the value.

			if (attr.avi == null) {
				MsgOut.println("Null attribute assignment");
			} else if (!processFirstOrderAttributes(attr)) {
				Attribute attrVar = getAttribute(attr.id);

				if (attrVar == null) {
					MsgOut.println("Missing attribute '" + attr.toString() + "'");
				} else {
					if (attrVar.cloneOnAssign) {
						attrVar = (Attribute) attrVar.clone();
						attrVar.cloneOnAssign = false;
						replaceAttribute(attrVar);
					}
					attrVar.avi = attr.avi;
				}
			}
			attr = attr.next;
		}
	} 

	public void assignRelnAttributes(RelationClass rc, EntityInstance dst, Attribute attr)
	{	
		// Find the associated edge
		System.out.println("Asgn: " + rc.getId() + " " + id + " " + dst.getId());
	}


/*
		v6.0.3
		Stupid to rescale -- might as well recalculate 
 */

	public void validateAllMyEdges() 
	{
		Enumeration			en;
		RelationInstance	ri;

		for (en = srcLiftedRelationElements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();
			if (ri.m_drawSrc != this) {
				System.out.println("validateAllMyEdges() " + this + " drawsrc " + ri.m_drawSrc);
				continue;
			}
			ri.validate();
		}

		for (en = dstLiftedRelationElements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();
			if (ri.m_drawDst != this) {
				System.out.println("validateAllMyEdges() " + this + " drawdst " + ri.m_drawDst);
				continue;
			}
			ri.validate();
	}	}

	protected void rescale(EdgePoint[] pts) {

		if (pts != null) {
			for (int i = 0; i < pts.length; i++) {
				if (pts[i] != null) {
					pts[i].clearRc();
	}	}	}	}

	public void setBounds(int x, int y, int width, int height) 
	{
		super.setBounds(x, y, width, height);
		rescaleChildren();
	}

	// For debugging
	public void setBounds(Rectangle r)
	{
		setBounds(r.x, r.y, r.width, r.height);
	}

	public void setBothBounds(int parentDiagramX, int parentDiagramY, int x, int y, int width, int height) 
	{
		m_diagramX = parentDiagramX + x;
		m_diagramY = parentDiagramY + y;


		setBounds(x, y, width, height);
	}

	// For debugging
	public void setLocation(int x, int y)
	{
		setBounds(x, y, getWidth(), getHeight());
	}

	// For debugging
	public void setLocation(Point p)
	{
		setLocation(p.x, p.y);
	}

	// For debugging
	public void setSize(int width, int height)
	{
		setBounds(getX(), getY(), width, height);
	}		

	// For debugging
	public void setSize(Dimension d)
	{
		setSize(d.width, d.height);
	}
	
	// Global coordinate transform
	//
	// Calculate global values for Rectangle from local with scale.
	// The absolute x and y offsets of the parent are passed.
	//
	// This also handles calling the rescale on the containing entities.
	//

	public void rescale(int parentDiagramX, int parentDiagramY, int parent_width, int parent_height)
	{
		int 		new_x, new_y, new_width, new_height;

		new_x      = (int) (parent_width  * m_xrelLocal);
		new_y      = (int) (parent_height * m_yrelLocal);
		new_width  = (int) (parent_width  * m_widthrelLocal);
		new_height = (int) (parent_height * m_heightrelLocal);

/*		System.out.println("EntityInstance.rescale(" + parent_width + ", " + parent_height + ") using " + this + " [" + m_xrelLocal + ", " + m_yrelLocal + ", " + m_widthrelLocal + ", " + m_heightrelLocal + "] = {" + new_x + ", " + new_y + ", " + new_width + ", " + new_height + "}");
		java.lang.Thread.dumpStack();
		System.out.println("-----");
*/
		// This set bounds causes children to be rescaled
		setBothBounds(parentDiagramX, parentDiagramY, new_x, new_y, new_width, new_height);
	}

	public void rescaleChildren() 
	{
		Enumeration		en;
		int				width, height;
		EntityInstance	e;
		
		width  = getWidth();
		height = getHeight();
		
		// rescale edge points

		rescale(topPoints);
		rescale(bottomPoints);
		rescale(leftPoints);
		rescale(rightPoints);
		
		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement();
			e.rescale(m_diagramX, m_diagramY, width, height);
		}
	}

	public Rectangle getDiagramBounds() 
	{
		return new Rectangle(getDiagramX(), getDiagramY(), getWidth(), getHeight());
	}

	protected void scale(double wf, double hf, boolean doContainer, boolean doPos)
	{
		// Scale ourselves 

//		System.out.println("EntityInstance.scale(" + wf + "," + hf + "," + doContainer + "," + doPos + ")");

		if (doContainer) {
			double x1, y1, w1, h1; 

			x1 = xRelLocal();		
			y1 = yRelLocal();
			w1 = widthRelLocal()  * wf;
			h1 = heightRelLocal() * hf;

			if (doPos) {
				x1 *= wf;
				y1 *= hf;
			}

			if (x1 >= 0.0 && w1 > 0.0 && y1 >= 0.0 && h1 > 0.0 && (x1+w1) <= 1.0 && (y1+h1) <= 1.0) {
				setxRelLocal(x1);
				setyRelLocal(y1);
				setwidthRelLocal(w1);
				setheightRelLocal(h1);	
			}
			return;

		} 
		
		// Insure the scale of the bounding box of the contained elements
		// will be within the container.
		
		RelRectangle bb = getChildrenRelBoundingBox();

		if (bb == null) {
			return;
		}

		bb.width  *= wf;
		bb.height *= hf;

		if (doPos) {
			bb.x = bb.x * wf;
			bb.y = bb.y * hf;
		}

		// Don't bother, if it won't fit right
		if (bb.x >= 0.0 && bb.width > 0.0 && bb.y >= 0.0 && bb.height > 0.0 && (bb.x + bb.width) < 1.0 && (bb.y + bb.height) < 1.0) {
		
			Enumeration en;

			// ... and then our children

			for (en = getChildren(); en.hasMoreElements(); ) {
				EntityInstance e = (EntityInstance) en.nextElement();
				e.scale(wf, hf, true, true);
		}	}
	}

	public void scale(double sf, boolean doContainer, boolean doPos) 
	{
		scale(sf, sf, doContainer, doPos);
	}

	public void scale(double sf, boolean doContainer) 
	{
		scale(sf, sf, doContainer, false);
	}

	public void scale(double sf) 
	{
		scale(sf, sf, true, false);
	}

	public void scaleX(double sf, boolean doContainer, boolean doPos) 
	{
		scale(sf, 1.0, doContainer, doPos);
	}

	public void scaleX(double sf, boolean doContainer) 
	{
		scale(sf, 1.0, doContainer, false);
	}

	public void scaleX(double sf) 
	{
		scale(sf, 1.0, true, false);
	}

	public void scaleY(double sf, boolean doContainer, boolean doPos) 
	{
		scale(1.0, sf, doContainer, doPos);
	}

	public void scaleY(double sf, boolean doContainer) 
	{
		scale(1.0, sf, doContainer, false); 
	}

	public void scaleY(double sf) 
	{
		scale(1.0, sf, true, false);
	}

	// Attempt to fit children to us  

	private void doFitScale(double dx, double dy, double wf, double hf) 
	{
		setxRelLocal(xRelLocal() * wf + dx);
		setyRelLocal(yRelLocal() * hf - dy);
		setwidthRelLocal(dx + widthRelLocal() * wf);
		setheightRelLocal(dy + heightRelLocal() * hf);	
	}


	public void fitTo(boolean alwaysFit) 
	{
		Enumeration en;

		MsgOut.dprintln("Fit to: " + id );

		if ( (widthRelLocal() == 0) || (heightRelLocal() == 0)) {
			containsList = new Hashtable();
		}

		if (containsList.size() == 0) {
			MsgOut.dprintln("No children");
			return;
		}

		RelRectangle bb = getChildrenRelBoundingBox();

		if (bb.width <= 0.0 || bb.height <= 0.000) {
			return;
		}

		if (!alwaysFit) {

			// Only continue if bb isn't completely contained in this

			if (bb.x >= 0.0 && bb.y >= 0.0 && (bb.x + bb.width) <= 1.0 && (bb.y + bb.height) <= 1.0) {
				return;
		}	}
		

		// Translate bb to top left of parent and scale to fit

		// want new_relx = 1/15 and (new_relx + new_relwidth) = 14/15 for bounding box
		// so:
		//	m*x         + c =  1/15
		//	m*(x+width) + c = 14/15
		//	m*(x+width-x)   = 13/15 => m = 13/(15*width)
		//							=> c = (1/15) - m * x;

		double scaleX = 13.0 / (15.0 * bb.width);
		double scaleY = 13.0 / (15.0 * bb.height);
		double dx     = (1.0 / 15.0) - scaleX * bb.x;
		double dy     = (1.0 / 15.0) - scaleY * bb.y;

		for (en = getChildren(); en.hasMoreElements(); ) {
			EntityInstance e = (EntityInstance) en.nextElement();

			e.doFitScale(dx, dy, scaleX, scaleY);
		}

		MsgOut.dprintln("Fit: " + scaleX + " " + scaleY);
		return;
	}

	// Change the relative dimensions if anything is out of wack -- ie relative < 1/15 or > 14/15

	public void verifyFit() 
	{
		fitTo( false );
	}


	public void deltaFont(int delta) 
	{
		setFontDelta(m_fontDelta + delta);
	}

	// Overridden by BaseEntity

	protected void drawTopLeftLabel(Graphics g) 
	{
		g.setFont(smallFont);
//		g.setColor(Color.black); 
		Util.drawStringClipped(g, m_label, MARGIN, MARGIN, getWidth()-MARGIN*2, getHeight()-MARGIN*2);
	}

	public Dimension getLabelDim(Graphics g, int type, boolean wParent) 
	{
		setFont(g, type);

		FontMetrics fm = g.getFontMetrics();

		int h = Do.fontHeight(fm);
		int w = fm.stringWidth(m_label);

		if (wParent) {
			EntityInstance pe = getEnterableParent();

			if (pe != null) {
				w = Math.max(fm.stringWidth(pe.getLabel() + " | "), w);
				h += h/2;
		}	} 
		return new Dimension(w, h);
	}



	public Dimension getLabelDim(Graphics g, int type) 
	{
		return getLabelDim(g, type, false);
	}



	public Dimension getFitDim(Graphics g, int type, boolean wParent) 
	{
		Dimension dim = getLabelDim(g, type, wParent);

		if (hasChildren() && !isOpen()) {
			dim.width += CONTENTS_FLAG_DIM;
		}
		dim.width  += MARGIN*3;
		dim.height += (MARGIN*3)/2;

		return dim;
	}

	public Dimension getFitDim(Graphics g, int type) 
	{
		return getFitDim(g, type, false);
	}


	// Edges

	public void invalidateAllEdges() 
	{ 
		Enumeration		 en;
		RelationInstance ri;

//		System.out.println("EntityInstance.drawAllEdges " + this);
		
		for (en = srcLiftedRelationElements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();
			ri.invalidateEdge();
		}	

		for (en = dstLiftedRelationElements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();
			ri.invalidateEdge();
	}	}	

	public void drawAllEdges(EntityInstance under) 
	{ 
		Enumeration		 en;
		EntityInstance	 src;
		RelationInstance ri;

//		System.out.println("EntityInstance.drawAllEdges " + this);
		
		for (en = srcLiftedRelationElements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();
			ri.draw(true);	// allow elision from source
		}	

		for (en = dstLiftedRelationElements(); en.hasMoreElements(); ) {
			ri  = (RelationInstance) en.nextElement();
			src = ri.m_drawSrc;
			if (!under.hasDescendant(src)) {
				ri.draw(true);	// allow elision to destination
		}	}	

//		System.out.println("EntityInstance.drawAllEdges " + this + " done");
	}

	public void drawHighlightedEdges() 
	{ 
		Enumeration		 en;
		RelationInstance ri;

		for (en = srcLiftedRelationElements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();
			ri.drawHighlighted();  
		}	

		for (en = dstLiftedRelationElements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();
			ri.drawHighlighted(); 
		}	
	}

	public void getHighlightedEdges(Vector v) 
	{ 
		Enumeration en;
		RelationInstance ri;
		EntityInstance root, src, e;


		for (en = m_srcRelList.elements(); en.hasMoreElements();) {
			ri = (RelationInstance) en.nextElement();
			if (ri.getHighlightFlag() && ri.getRelationClass().isActive()) {
				v.addElement(ri);
			}
		}

		root = dg.getDrawRoot();
		for (en = m_dstRelList.elements(); en.hasMoreElements(); ) {
			ri  = (RelationInstance) en.nextElement();
			src = ri.getSrc();

			if (!root.hasDescendant(src) && ri.getHighlightFlag() && ri.getRelationClass().isActive()) {
				v.addElement(ri);
			}
		}


		// Call method on each child

		for (en = getChildren(); en.hasMoreElements();) {
			e = (EntityInstance) en.nextElement();
			e.getHighlightedEdges(v);
		}
	}


	// TA raw info routines

	public void getInstancesRaw(Vector v, Hashtable st) 
	{
		Enumeration en;
		EntityInstance child;

		// Get instance info for ourself, and then all our children
		if (m_parentClass != null) {
			// Not $ROOT 
			Integer[] pair = new Integer[2];

			pair[0] = (Integer) st.get(id);
			pair[1] = (Integer) st.get(m_parentClass.getId());
			v.addElement(pair);
		}
		for (en = getChildren(); en.hasMoreElements(); ) {
			child = (EntityInstance) en.nextElement();
			child.getInstancesRaw(v, st);
		}
	}

	public void getRelationsRaw(Vector v, Hashtable st, Integer cid) 
	{
		// Get the relations on us, and then all our children

		Enumeration en; 

		for (en = m_srcRelList.elements(); en.hasMoreElements(); ) {
			RelationInstance ri = (RelationInstance) en.nextElement();
			Integer[] triple = new Integer[3];
			RelationClass rc  = ri.getRelationClass();
			EntityInstance e1 = ri.getSrc();
			EntityInstance e2 = ri.getDst();
			triple[0] = (Integer) st.get(rc.getId());
			triple[1] = (Integer) st.get(e1.getId());
			triple[2] = (Integer) st.get(e2.getId());

			v.addElement(triple);
		}				
		for (en = getChildren(); en.hasMoreElements(); ) {
			EntityInstance child = (EntityInstance) en.nextElement();
			if (m_parentClass != null) {
				// If not root, output contain relation
				Integer[] triple = new Integer[3];
				triple[0] = cid;
				triple[1] = (Integer) st.get(id);
				triple[2] = (Integer) st.get(child.id);
				v.addElement(triple);
			}
			child.getRelationsRaw(v, st, cid);
		}
	}

	// TA file output routines 

	public void writeInstances(PrintStream ps) throws IOException 
	{
		// Write an instance line for ourself, and then all our children

		if (m_parentClass != null) {
			// Not $ROOT 
			ps.print(Diagram.INSTANCE_ID + " " + qt(id) + " " + qt(m_parentClass.getId()) + "\n");
		}

		Enumeration en;

		for (en = getChildren(); en.hasMoreElements(); ) {
			EntityInstance child = (EntityInstance) en.nextElement();
			child.writeInstances(ps);
		}
	}

	public void writeRelations(PrintStream ps) throws IOException {

		// Write out the relations on us, and then all our children

		Enumeration en; 

		for (en = m_srcRelList.elements(); en.hasMoreElements(); ) {
			RelationInstance ri = (RelationInstance) en.nextElement();
			ri.writeRelation(ps); 
		}				

		for (en = getChildren(); en.hasMoreElements(); ) {
			EntityInstance child = (EntityInstance) en.nextElement();
			if (m_parentClass != null) {
				// If not root, output contain relation
				ps.print(Diagram.CONTAIN_ID + " " + qt(id) + " " + qt(child.id) + "\n"); 
			}
			child.writeRelations(ps);
		}
	}

	protected void writeLayout(PrintStream ps) throws IOException 
	{
		// Write out Rectangle attributes

		ps.print(Attribute.indent + "xrel      = " + xRelLocal() + "\n");
		ps.print(Attribute.indent + "yrel      = " + yRelLocal() + "\n"); 
		ps.print(Attribute.indent + "widthrel  = " + widthRelLocal() + "\n"); 
		ps.print(Attribute.indent + "heightrel = " + heightRelLocal() + "\n");
	}

	protected void writeElision(PrintStream ps, Vector ev, String id) 
	{
		if (ev.size() > 0) {
			ps.print(Attribute.indent + id + " = (");
			Enumeration en = ev.elements();
			while(en.hasMoreElements()) {
				String rcId = (String) en.nextElement(); 
				ps.print(rcId);
				if (en.hasMoreElements()) {
					ps.print(" ");
			}	}
			ps.print(")\n"); 
		}
	}

	protected void writePoints(PrintStream ps, EdgePoint[] pts, String name) 
	{
		if (pts == null) {
			return;
		}

		boolean isDefault = true;

		for (int i = 1; i < pts.length; i++) {
			if (pts[i] != null) {
				isDefault = isDefault && pts[i].isDefault;
		}	}

		if (isDefault) {
			return;
		}

		ps.print("	" + name + " = ( ");

		for (int i = 1; i < pts.length; i++) {
			if (pts[i] != null && !pts[i].isDefault) {
				ps.print("(" + pts[i].getRc().getId() + " " + pts[i].m_wf + " " + pts[i].m_hf + ") ");
			}
		}
		ps.print(")\n");
	}

	protected void regRawElision(Vector v, Hashtable st, Vector ev, String id) 
	{
		if (ev.size() > 0) {
			int len = ev.size();
			String[] sl = new String[len];

			for (int i=0; i<len; i++) {
				sl[i] = (String) ev.elementAt(i);
			}
			regRawAttribute(v, st, id, sl);
		}
	}

	public void	writeAttributes(PrintStream ps) throws IOException
	{  
		Enumeration en;

		// Write out attribute record for us, and then our children

		ps.print(qt(id) + " {\n");
		writeLayout(ps);

		if (m_label != null && !m_label.equals(mungeId(id))) {
			if (m_parentClass == null || !m_parentClass.defaultValue(LABEL_ID, m_label)) {
				// Save only if it differs from munged id (ie. auto-label)
				ps.print(Attribute.indent + LABEL_ID + " = " + qt(m_label) + "\n");
		}	}

		if (description != null) {
			if (m_parentClass == null || !m_parentClass.defaultValue(DESC_ID, description)) {
				ps.print(Attribute.indent + DESC_ID + " = \"" + description + "\"\n");
		}	}

		if (m_parentClass != null) {
			// First order attributes not associated with $ROOT 
			writeElision(ps, m_dstElision, IN_ELISION_ID);
			writeElision(ps, m_srcElision, OUT_ELISION_ID);
			writeElision(ps, m_enteringElision, CLIENT_ELISION_ID);
			writeElision(ps, m_exitingElision, SUPPLIER_ELISION_ID);
			writeElision(ps, m_internalElision, INTERNAL_ELISION_ID);
			writePoints(ps, topPoints, INPOINT_ID);
			writePoints(ps, bottomPoints, OUTPOINT_ID);
			writePoints(ps, leftPoints, LEFTPOINT_ID);
			writePoints(ps, rightPoints, RIGHTPOINT_ID);

			if (m_fontDelta != 0) {
				ps.print(Attribute.indent + FONTDELTA_ID + " = ");
				ps.print(m_fontDelta + "\n");
			}
		}

		// Finally output the second-class attributes 

		super.writeAttributes(ps, m_parentClass, false); 

		// End the record 

		ps.print("}\n"); 

		// Write any attributes of src relations

		for (en = m_srcRelList.elements(); en.hasMoreElements(); ) {
			RelationInstance ri = (RelationInstance) en.nextElement();
			ri.writeAttributes(ps);
		}

		// Recurse for contained children

		for (en = getChildren(); en.hasMoreElements(); ) {
			EntityInstance child = (EntityInstance) en.nextElement();
			child.writeAttributes(ps);
		}
	}

	public void getAttributesRaw(Vector mv, Hashtable st) {

		Vector v = new Vector();

 		regRawAttribute(v, st, XRELPOSITION_ID, m_xrelLocal);
		regRawAttribute(v, st, YRELPOSITION_ID, m_yrelLocal);
		regRawAttribute(v, st, WIDTHREL_ID, m_widthrelLocal);
		regRawAttribute(v, st, HEIGHTREL_ID, m_heightrelLocal);

		if (m_label != null && !m_label.equals(mungeId(id))) {
			// Save only if it differs from munged id (ie. auto-label)
			regRawAttribute(v, st, LABEL_ID, m_label);
		}

		if (description != null) {
			regRawAttribute(v, st, DESC_ID, description);
		}

		if (m_parentClass != null) {
			// First order attributes not associated with $ROOT 
			Color pc	 = m_parentClass.getObjectColor();
			Color color1 = getObjectColor();
			if (!color1.equals(pc)) {
				regRawAttribute(v, st, COLOUR_ID, color1);
			}
			pc = m_parentClass.getLabelColor();
			Color labelcolor1;
			labelcolor1 = getLabelColor();
			if (!labelcolor1.equals(pc)) {
				regRawAttribute(v, st, LABEL_COLOUR_ID, labelcolor1);
			}
			color1 = getObjectColorWhenOpen();
			if (color1 != null) {
				regRawAttribute(v, st, OPEN_COLOUR_ID, labelcolor1);
			}
			regRawElision(v, st, m_dstElision, IN_ELISION_ID);
			regRawElision(v, st, m_enteringElision, OUT_ELISION_ID);
			regRawElision(v, st, m_exitingElision, SUPPLIER_ELISION_ID);
			regRawElision(v, st, m_internalElision, INTERNAL_ELISION_ID);

			// addRawIOpoints(v, st);
			if (m_fontDelta != 0) {
				regRawAttribute(v, st, FONTDELTA_ID, m_fontDelta);
			}
		}

		// Finally output the second-class attributes 
		if (attributes.size() > 0) 
			super.getAttributesRaw(v, st, false);

		Object[] obj = new Object[2];

		obj[0] = st.get(id);
		obj[1] = v;

		mv.addElement(obj);

		Enumeration en;

/*
		// Write any attributes of src relations

		Enumeration en = m_srcRelList.elements();

		while(en.hasMoreElements()) {
			RelationInstance ri = (RelationInstance) en.nextElement();
			ri.writeAttributes(ps);
		}
*/

		// Recurse for contained children

		en = getChildren();

		while(en.hasMoreElements()) {
			EntityInstance child = (EntityInstance) en.nextElement();
			child.getAttributesRaw(mv, st);
		}
	}

	// Accessor functions and queries 



	public boolean canOpen() 
	{
		return !containsList.isEmpty();
	}

	public void setOpen() 
	{
		if (!isMarked(OPEN_MARK)) {
			nandMark(CLOSED_MARK);
			orMark(OPEN_MARK);
	}	}

	public void setClosed()
	{
		if (!isMarked(CLOSED_MARK)) {
			nandMark(OPEN_MARK);
			orMark(CLOSED_MARK);
	}	}

	public void openStatusUnknown()
	{
		nandMark(OPEN_MARK|CLOSED_MARK);
	}

	public boolean isOpen() 
	{
		// If haven't yet decided if open or closed
		// We will treat zero sized objects as being closed so that we don't have to repaint edges to zero
		// sized objects which is very expensive if have huge tree and everything open

		if (!isMarked(OPEN_MARK|CLOSED_MARK)) {
			if (canOpen() && getWidth() > 0 && getHeight() > 0 && !isDstRelationElided(Diagram.CONTAIN_ID)) {
				setOpen();
			} else {
				setClosed();
		}	}

		return(isMarked(OPEN_MARK));
	}

	public EntityInstance getContainedBy() 
	{ 
		return m_containedBy;
	}	

	public EntityInstance getEnterableParent() 
	{
		EntityInstance e;
		
		for (e = m_containedBy; e != null; e = e.m_containedBy) {
			if (e.isEnterable()) {
				break;
		}	}
		return e;
	}



	public EntityClass getEntityClass() 
	{
		return m_parentClass;
	}

	private void setParentClass(EntityClass parentClass)
	{
/*
		System.out.println("EntityInstance:setParentClass(" + parentClass + ") from " + m_parentClass + " for " + getLabel());		// IJD
		java.lang.Thread.dumpStack();
		System.out.println("-----");
*/
		m_parentClass = parentClass;
	}

	public boolean isRoot() {

		return (m_containedBy == null);

	}



	public boolean hasChildren() 
	{
		return !containsList.isEmpty();
	}



	public int numChildren() 
	{
		return containsList.size();
	}

	public Enumeration getChildren() 
	{
		return containsList.elements();
	}


	public String toString() 
	{
		return m_label;
	}

	public Diagram getDiagram() 
	{
		return dg;
	}

	public String getDescription() 
	{
		if (description != null)
			return description; 
		else
			return "";	
	}

	public void setDescription(String str)
	{
		description = str;
	}

	public void setLabel(String str) 
	{
		m_label = str;
	}

	// Rectangle is global, adjust it to coords local to parent

	public void setDiagramBounds(int diagramX, int diagramY, int width, int height) 
	{
		double	width1, height1;

		m_diagramX = diagramX;
		m_diagramY = diagramY;
		width1     = parentsWidth();
		height1    = parentsHeight();

		if (width1 <= 0.0) {
			setxRelLocal(0);
			setwidthRelLocal(0);
		} else {
			setxRelLocal((diagramX - getParentDiagramX())/width1);
			setwidthRelLocal(width/width1);
		}

		if (height1 <= 0.0) {
			setyRelLocal(0);
			setheightRelLocal(0);
		} else {
			setyRelLocal((diagramY - getParentDiagramY())/height1);
			setheightRelLocal(height/height1);
		}
/*
		if (m_label.equals("code")) {
//			System.out.println("EntityInstance.setDiagramBounds [" + diagramX + "," + diagramY + "," + width + "," + height + "]");
			m_debug_x = (int) (diagramX-getParentDiagramX());
			m_debug_y = (int) (diagramY-getParentDiagramY());
			m_debug_width  = width;
			m_debug_height = height;
		}
 */	
		
//		System.out.println("EntityInstance.setDiagramBounds " + label + " xrel=" + m_xrelLocal + " yrel=" + m_yrelLocal + " widthrel=" + m_widthrelLocal + " heightrel=" + m_heightrelLocal );
		setBounds((int) (diagramX-getParentDiagramX()), (int) (diagramY-getParentDiagramY()), width, height);

		validateAllMyEdges();
		moveCardinals();
	} 

	public void setDiagramBounds(Rectangle lyt) 
	{
		setDiagramBounds(lyt.x, lyt.y, lyt.width, lyt.height);
	} 

	public Rectangle getChildBoundingBox() 
	{
		Enumeration en;
		EntityInstance e;

		// Calc the global bounding box of children

		if (containsList.isEmpty()) {
			return null;
		}

		int x1 = Integer.MAX_VALUE;
		int x2 = Integer.MIN_VALUE;
		int y1 = Integer.MAX_VALUE;
		int y2 = Integer.MIN_VALUE;

		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement();
			if (e.getX() < x1) {
				x1 = e.getX();
			}
			if (e.getY() < y1) {
				y1 = e.getY();
			}
			if (e.getX() + e.getWidth() > x2) {
				x2 = e.getX() + e.getWidth();
			}
			if (e.getY() + e.getHeight() > y2) {
				y2 = e.getY() + e.getHeight();
		}	}
		return new Rectangle(x1, y1, x2-x1, y2-y1);
	}

	public Color getBackgroundWhenOpen() {

		Color c;

		// Container is handled specially
		// Find colour from depth

		c = getObjectColorWhenOpen();
		if (c == null) {

			float v = Diagram.BG - 0.05f;

			EntityInstance b = this;

			if (isTransparent()) {
				b = b.m_containedBy;
			}

			EntityInstance root = dg.getDrawRoot();

			if (!root.hasDescendant(this)) {
				root = root.commonAncestor(this);
			}


			while(b != root) {
				v -= 0.05f;
				b = b.m_containedBy; 
			}

			c = new Color(v, v, v);

			if (Util.isBlack(c)) {
				c = Color.lightGray;
		}	}
		return c;
	}
			

	public Color getBackground() 
	{
		Color c;

		if (isOpen()) {
			c = getBackgroundWhenOpen();
		} else {
			c = getObjectColor();
		}
		return c;
	}



	public boolean isRelationElided(Vector v, String id) 
	{
		return v.contains(id);
	}



	protected boolean toggleElision(Vector v, String id) 
	{
		if (isRelationElided(v, id)) {
			if (!v.removeElement(id)) {
				MsgOut.println("remove() failed");
			}
			return false; 
		} 

		v.addElement(id); 
		return true; 
	}

 

	public boolean toggleDstElision(String id) 
	{
		MsgOut.dprintln("toggle target " + id);
		return toggleElision(m_dstElision, id);
	}

 

	public boolean toggleDstElision(RelationClass rc) 
	{
		return toggleDstElision(rc.getId());
	}



	public boolean toggleSrcElision(String id) 
	{
		MsgOut.dprintln("toggle out " + id);
		return toggleElision(m_srcElision, id);
	}



	public boolean toggleSrcElision(RelationClass rc)
	{
		return toggleSrcElision(rc.getId());
	}



	public boolean toggleEnteringElision(String id) 
	{
		MsgOut.dprintln("toggle client " + id);
		return toggleElision(m_enteringElision, id);
	}



	public boolean toggleEnteringElision(RelationClass rc) 
	{
		return toggleEnteringElision(rc.getId());
	}



	public boolean toggleExitingElision(String id) {

		MsgOut.dprintln("toggle exiting " + id);



		return toggleElision(m_exitingElision, id);

	}



	public boolean toggleExitingElision(RelationClass rc) 
	{
		return toggleExitingElision(rc.getId());
	}



	public boolean toggleInternalElision(String id) 
	{
		MsgOut.dprintln("toggle internal " + id);
		return toggleElision(m_internalElision, id);
	}



	public boolean toggleInternalElision(RelationClass rc) 
	{
		return toggleInternalElision(rc.getId());
	}

	// Queries 



	public boolean hasId(String id) 
	{
		return this.id.equals(id);
	}



	public boolean isDstRelationElided(RelationClass rc) 
	{
		return isRelationElided( m_dstElision, rc.getId() );
	}



	public boolean isDstRelationElided(String id) 
	{
		return isRelationElided(m_dstElision, id);
	}

	public boolean isSrcRelationElided(RelationClass rc) 
	{
		return isRelationElided(m_srcElision, rc.getId());
	}

	public boolean isSrcRelationElided(String id) 
	{
		return isRelationElided(m_srcElision, id);
	}



	public boolean isEnteringRelationElided(RelationClass rc) 
	{
		return isRelationElided(m_enteringElision, rc.getId());
	}



	public boolean isEnteringRelationElided(String id) 
	{
		return isRelationElided(m_enteringElision, id);
	}



	public boolean isExitingRelationElided(RelationClass rc) 
	{
		return isRelationElided(m_exitingElision, rc.getId());
	}



	public boolean isExitingRelationElided(String id) 
	{
		return isRelationElided(m_exitingElision, id);
	}



	public boolean isInternalRelationElided(RelationClass rc) 
	{
		return isRelationElided(m_internalElision, rc.getId());
	}

	public boolean isInternalRelationElided(String id) 
	{
		return isRelationElided(m_internalElision, id);
	}

	public int getPostorder()
	{
		return (m_postorder);
	}

	public boolean hasAncestor(EntityInstance e) 
	{
		if (m_preorder == 0 || e.m_preorder == 0 || m_postorder == 0 || e.m_postorder == 0) {
			System.out.println("EntityInstance.hasAncestor error " + this + "(" + m_preorder + "," + m_postorder + ") " + e + "(" + e.m_preorder + "," + e.m_postorder + ")");
		}

		return(e.m_preorder < m_preorder && m_postorder < e.m_postorder);
	}


	// Returns true if e is a descendant of me

	public boolean hasDescendant(EntityInstance e) 
	{
		if (m_preorder == 0 || e.m_preorder == 0 || m_postorder == 0 || e.m_postorder == 0) {
			System.out.println("EntityInstance.hasDescendant error " + this + "(" + m_preorder + "," + m_postorder + ") " + e + "(" + e.m_preorder + "," + e.m_postorder + ")");
		}
		return(m_preorder < e.m_preorder && e.m_postorder < m_postorder);
	}

	public boolean hasDescendantOrSelf(EntityInstance e) 
	{
		if (m_preorder == 0 || e.m_preorder == 0 || m_postorder == 0 || e.m_postorder == 0) {
			System.out.println("EntityInstance.hasDescendant error " + this + "(" + m_preorder + "," + m_postorder + ") " + e + "(" + e.m_preorder + "," + e.m_postorder + ")");
		}
		return(m_preorder <= e.m_preorder && e.m_postorder <= m_postorder);
	}

	public boolean isPathBetween(EntityInstance e)
	{
		if (m_preorder == 0 || e.m_preorder == 0 || m_postorder == 0 || e.m_postorder == 0) {
			System.out.println("EntityInstance.isPathBetween error " + this + "(" + m_preorder + "," + m_postorder + ") " + e + "(" + e.m_preorder + "," + e.m_postorder + ")");
		}
		return((m_preorder <= e.m_preorder && e.m_postorder <= m_postorder) || (e.m_preorder < m_preorder && m_postorder < e.m_postorder));
	}

	// Return the common ancestor entity

	public EntityInstance commonAncestor(EntityInstance e) 
	{
		EntityInstance e1;

		for (e1 = m_containedBy; e1 != null; e1 = e1.m_containedBy) {
			if (e1.hasDescendant(e)) {
				break;
		}	}
		return(e1);
	}

	// Finds that entity being displayed either as a client or a server
	// to which arrows to my entity should be addressed

	public EntityInstance clientServer() 
	{
		EntityInstance e1;

		for (e1 = this; e1 != null; e1 = e1.m_containedBy) {
			if (e1.isMarked(CLIENT_MARK | SUPPLIER_MARK)) {
				break;
		}	}
		return(e1);
	}

	public boolean inDiagram()
	{
		return(isMarked(IN_DIAGRAM));
	}

	public EntityInstance getVisibleEntity() 
	{
		EntityInstance e;

		for (e = this; e != null; e = e.m_containedBy) {
			if (inDiagram()) {
				break;
		}	}
		return e;
	}

	// Return true if we're a client of passed entity.
	// That is, a relation exist between us, or our children and 
	// the passed entity or its children.

	public boolean isClientOf(EntityInstance pc) 
	{
		Enumeration en;
		
		for (en = m_srcRelList.elements(); en.hasMoreElements(); ) {
			RelationInstance ri = (RelationInstance) en.nextElement();
			EntityInstance e = ri.getDst();

			if (e == pc) {
				return true;
			}

			if (pc.hasAncestor(e)) {
				return true;
			}
		}

		for (en = getChildren(); en.hasMoreElements(); ) {
			EntityInstance e = (EntityInstance) en.nextElement();
			if (e.isClientOf(pc)) {
				return true;
		}	}

		return false;
	}

	// Return reference to entity the cursor is over. 
	// Return null if none.

	public EntityInstance getMouseOver(int x, int y) 
	{
		if (contains(x, y)) {
			if (isOpen()) {
				Enumeration en = getChildren();

				while(en.hasMoreElements()) {
					EntityInstance e = (EntityInstance) en.nextElement();

					EntityInstance over = e.getMouseOver(x, y);
					if (over != null)
						return over; 
				}

			}
			// It's not over any of our children, so it's over us.
			return this; 
		}
		return null;
	}

	public EntityInstance intersects(Rectangle lyt) 
	{
		if (containsRectangle(lyt)) {
			if (!containsList.isEmpty()) {
				// Check children
				Enumeration en;

				for (en = getChildren(); en.hasMoreElements(); ) {
					EntityInstance e = (EntityInstance) en.nextElement();
					EntityInstance oe = e.intersects(lyt); 
					if (oe != null) {
						return oe; 
				}	}
				return null; 
			}
			return this;  // over us 
		} else if (intersectsLayout(lyt)) { 
			return this;
		}
		return null; 
	}

	public EntityInstance containing(Rectangle lyt) 
	{
		if (containsRectangle(lyt)) {
			if (!containsList.isEmpty() && isOpen()) {
				// Check children
				Enumeration en;

				for (en = getChildren(); en.hasMoreElements(); ) {
					EntityInstance e = (EntityInstance) en.nextElement();
					EntityInstance oe = e.containing(lyt);
					if (oe != null) {
						return oe;
				}	}
			}
			return this;
		}
		return null;
	}

	public Enumeration srcRelationElements() 
	{
		return m_srcRelList.elements(); 
	}

	public Enumeration dstRelationElements() 
	{
		return m_dstRelList.elements(); 
	}

	public Enumeration srcLiftedRelationElements() 
	{
		return m_srcLiftedList.elements(); 
	}

	public Enumeration dstLiftedRelationElements() 
	{
		return m_dstLiftedList.elements(); 
	}

	protected void addSrcRels(Vector v) 
	{
		Enumeration en;
		EntityInstance e;

		for (en = m_srcRelList.elements(); en.hasMoreElements(); ) {
			v.addElement(en.nextElement());
		}

		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement();
			e.addSrcRels(v);
		}
	}

	public Enumeration srcRelationAbstractedElements()
	{
		// Return enumeration for all edges which
		// have this entity as the source (real and abstracted).

		Vector v = new Vector();
		addSrcRels(v);
		return v.elements();
	}

	protected void addDstRels(Vector v) 
	{
		Enumeration en;
		EntityInstance e;

		for (en = m_dstRelList.elements(); en.hasMoreElements(); ) {
			v.addElement(en.nextElement());
		}

		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement();
			e.addDstRels(v);
		}
	}

	public Enumeration dstRelationAbstractedElements() 
	{
		// Return enumeration for all edges which
		// have this entity as the destination (real and abstracted).

		Vector v = new Vector();
		addDstRels(v);
		return v.elements();
	}

	public Vector groupRegion(Rectangle lyt) 
	{
		// This entity isn't in the group, but any children which
		// intersect the passed region are.

		if (containsList.isEmpty()) {
			return null;
		}
		Vector grp = new Vector();
		Enumeration en;
		EntityInstance e;
		
		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement(); 
			if (e.intersectsLayout(lyt)) {
				e.setGroupFlag(); 
				grp.addElement(e);
			}
		}

		if (grp.isEmpty()) {
			return null;
		}
		return grp;
	}

	public void getGroup(Vector grp) 
	{
		Enumeration en;
		EntityInstance e;
		
		if (getGroupFlag()) {
			grp.addElement(this);
		}

		for (en = getChildren(); en.hasMoreElements();) {
			e = (EntityInstance) en.nextElement(); 
			e.getGroup(grp);
	}	}

	protected void setGroupFlag() 
	{
		if (!isMarked(GROUP_MARK)) {
			orMark(GROUP_MARK);
			repaint(); 
	}	}

	protected void clearGroupFlag() 
	{
		if (isMarked(GROUP_MARK|GROUPKEY_MARK)) {
			nandMark(GROUP_MARK|GROUPKEY_MARK);
			repaint();
	}	}		

	public boolean getGroupFlag() 
	{
		return isMarked(GROUP_MARK);
	}

	protected void setGroupKeyFlag() 
	{
		if (!isMarked(GROUPKEY_MARK)) {
			orMark(GROUPKEY_MARK);
			repaint();
	}	}

	protected void clearGroupKeyFlag() 
	{
		if (isMarked(GROUPKEY_MARK)) {
			nandMark(GROUPKEY_MARK);
			repaint();
	}	}

	public boolean getGroupKeyFlag() 
	{
		return isMarked(GROUPKEY_MARK);
	}

	public void getRedBoxGroup(Vector grp) 
	{
		Enumeration		en;
		EntityInstance	e;

		if (isMarked(REDBOX_MARK)) {
			grp.addElement(this);
		}

		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement(); 
			e.getRedBoxGroup(grp);
	}	}

	public void setRedBoxFlag() 
	{
		if (!isMarked(REDBOX_MARK)) {
			orMark(REDBOX_MARK);
			repaint();
	}	}

	public void clearRedBoxFlag() 
	{
		if (isMarked(REDBOX_MARK)) {
			nandMark(REDBOX_MARK);
			repaint();
	}	}

	public void clearAllFlags() 
	{
		Enumeration en;

		clearGroupFlag();
		clearGroupKeyFlag();
		clearRedBoxFlag();
		openStatusUnknown();

		for (en = m_srcRelList.elements(); en.hasMoreElements();) {
			RelationInstance ri = (RelationInstance) en.nextElement();
			ri.clearHighlightFlag();
			ri.clearGroupFlag();
		}
		for (en = getChildren(); en.hasMoreElements(); ) {
			EntityInstance e = (EntityInstance) en.nextElement(); 
			e.clearAllFlags();
		}
	}

	// Remove redbox and highlight flags at and under me

	public void clearQueryFlags() 
	{
		Enumeration en;
		RelationInstance ri;
		EntityInstance e;

		clearRedBoxFlag();

		for (en = m_srcRelList.elements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();
			ri.clearHighlightFlag();
		}

		for (en = getChildren(); en.hasMoreElements(); ) {
			e    = (EntityInstance) en.nextElement(); 
			e.clearQueryFlags();
	}	}

	public void clearGroupFlags() 
	{
		Enumeration en;
		EntityInstance e;

		clearGroupFlag();
		clearGroupKeyFlag();

		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement(); 
			e.clearGroupFlags();
		}
	}

	public boolean isClippedLabel() 
	{
		if (getWidth() < MIN_WIDTH || getHeight() < MIN_HEIGHT) {
			return(true);
		}
		return(false);
	}

	// Internal attributes of different entities

	public boolean isTransparent() 
	{
		return false;
	}

	public boolean isCloseable() 
	{
		return true;
	}

	public boolean isClickable() 
	{
		return true;
	}

	public boolean isLabelDrawable() 
	{
		return true;
	}

	public boolean isEnterable() 
	{
		return true;
	}

	public void orMark(int val) 
	{
/*
		if ((m_mark & REDBOX_MARK) == 0) {
			if ((val & REDBOX_MARK) != 0) {
				System.out.println("EntityInstance.orMark(REDBOX_MARK) " + this);
		}	}
*/
		m_mark |= val;
	}

	public void andMark(int val)
	{
/*
		if ((m_mark & REDBOX_MARK) != 0) {
			if ((val & REDBOX_MARK) == 0) {
				System.out.println("EntityInstance.andMark(~REDBOX_MARK) " + this);
		}	}
*/	
		m_mark &= val;
	}

	public void nandMark(int val) 
	{
/*
		if ((m_mark & REDBOX_MARK) != 0) {
			if ((val & REDBOX_MARK) != 0) {
				System.out.println("EntityInstance.nandMark(REDBOX_MARK) " + this);
		}	}	
*/
		m_mark &= ~val;
	}
	
	public boolean isMarked(int val)
	{
		return((m_mark & val) != 0);
	}

	public int getMark()
	{
		return m_mark;
	}

	public boolean isClient()
	{
		return(isMarked(CLIENT_MARK));
	}

	public boolean isSupplier()
	{
		return(isMarked(SUPPLIER_MARK));
	}

	public boolean isClientOrSupplier()
	{
		return(isMarked(CLIENT_MARK|SUPPLIER_MARK));
	}

	static public Font getSmallFont() 
	{
		return smallFont;
	}

	public double parentsWidthLocal()
	{
		if (m_containedBy == null) {
			if (dg == null) {
				System.out.println("Null diagram\n");
			}
			return(dg.getWidth());
		} 
		return(m_containedBy.parentsWidthLocal() * m_containedBy.widthRelLocal());
	}

	public double parentsHeightLocal()
	{
		if (m_containedBy == null) {
			return(dg.getHeight());
		} 
		return(m_containedBy.parentsHeightLocal() * m_containedBy.heightRelLocal());
	}

	public double parentsWidth()
	{
		if (m_containedBy == null) {
			if (dg == null) {
				System.out.println("Null diagram\n");
			}
			return(dg.getWidth());
		} 
		return(m_containedBy.getWidth());
	}

	public double parentsHeight()
	{
		if (m_containedBy == null) {
			return(dg.getHeight());
		} 
		return(m_containedBy.getHeight());
	}

	/* Relative local coordinate system
     *
     * 0 same offset as top/left hand edge of parent
     * 1 same offset as bottom/right hand edge of parent
     */

	public double xRelLocal()  
	{
		return(m_xrelLocal);
	}

	public double yRelLocal() 
	{
		return(m_yrelLocal);
	}

	public double widthRelLocal() 
	{
		return(m_widthrelLocal);
	}

	public double heightRelLocal() 
	{
		return(m_heightrelLocal);
	}

	public void setxRelLocal(double xRelLocal)  
	{
		if (xRelLocal >= 0) {
			m_xrelLocal = xRelLocal;
		} else {
			m_xrelLocal = 0;
	}	}

	public void setyRelLocal(double yRelLocal) 
	{
		if (yRelLocal >= 0) {
			m_yrelLocal = yRelLocal;
		} else {
			m_yrelLocal = 0;
	}	}	

	public void setwidthRelLocal(double widthRelLocal) 
	{
		if (widthRelLocal > 1.0) {
			widthRelLocal = 1.0;
		}
		if ((m_xrelLocal + widthRelLocal) > 1.0) {
			m_xrelLocal     = 1.0 - widthRelLocal;
		}
		m_widthrelLocal = widthRelLocal;
	}

	public void setheightRelLocal(double heightRelLocal) 
	{
		if (heightRelLocal > 1.0) {
			heightRelLocal = 1.0;
		}
		if (m_yrelLocal + heightRelLocal > 1.0) {
			m_yrelLocal = 1.0 - heightRelLocal;
		}
		m_heightrelLocal = heightRelLocal;
	}

	/* Local coordinate system (obsoleted because of numeric underflow)
	 *
	 * 0 - Same as top/left edge of parent
	 * n - As position n while parent at position m.
     */

	/* This is only called if we are using legacy coordinate system..
	 * It is called after loading coordinates
	 * and converts the old local coordinates temporarily stored into relative ones
	 */

	public void computeRelCoordinates(double xParent, double yParent, double widthParent, double heightParent)
	{
		double			xLocal, yLocal, widthLocal, heightLocal;
		Enumeration		en;
		EntityInstance	e;

		xLocal      = m_xrelLocal;
		yLocal      = m_yrelLocal;
		widthLocal  = m_widthrelLocal;
		heightLocal = m_heightrelLocal;

		if (widthParent == 0.0) {
			m_xrelLocal      = 0.0;
			m_widthrelLocal  = 0.0;
		} else {
			m_xrelLocal      = xLocal/widthParent;
			m_widthrelLocal  = widthLocal/widthParent;
		}
		if (heightParent == 0.0) {
			m_yrelLocal      = 0.0;
			m_heightrelLocal = 0.0;
		} else {
			m_yrelLocal      = yLocal/heightParent;
			m_heightrelLocal = heightLocal/heightParent;
		}

		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement();
			e.computeRelCoordinates(xLocal, yLocal, widthLocal, heightLocal);
		}
//		System.out.println("EntityInstance:computeRelCoordinates(" + xParent + ", " + yParent + ", " + widthParent + ", " + heightParent + ") = {" + m_xrelLocal + ", " + m_yrelLocal + ", " + m_widthrelLocal + ", " + m_heightrelLocal + "}");
	}

	public void setxLocal(double xLocal)  
	{
		dg.m_uses_local_coordinates = true;
		m_xrelLocal                 = xLocal;
	}


	public void setyLocal(double yLocal) 
	{
		dg.m_uses_local_coordinates = true;
		m_yrelLocal                 = yLocal;
	}

	public void setwidthLocal(double widthLocal) 
	{
		dg.m_uses_local_coordinates = true;
		m_widthrelLocal             = widthLocal;
	}

	public void setheightLocal(double heightLocal) 
	{
		dg.m_uses_local_coordinates = true;
		m_heightrelLocal            = heightLocal;
	}

	public void	  ancilliary(Object value)
	{
		m_ancilliary = value;
	}

	public Object ancilliary()
	{
		return(m_ancilliary);
	}

	// The routines that follow hide the complexity of getting/setting attribute values
	// from EditAttributes
 
	public int getPrimaryAttributeCount()
	{
		return(13);
	}

	public String getAttributeNameAt(int index)
	{
		String	name;

		switch (index) {
		case 0:
			name  = "id";
			break;
		case 1:
			name  = "class";
			break;
		case 2:
			name  = LABEL_ID;
			break;
		case 3:
			name  = TITLE_ID;
			break;
		case 4:
			name  = DESC_ID;
			break;	
		case 5:
			name  = COLOUR_ID;
			break;
		case 6:
			name  = LABEL_COLOUR_ID;
			break;
		case 7:
			name  = OPEN_COLOUR_ID;
			break;
		case 8:
			name  = XRELPOSITION_ID;
			break;
		case 9:
			name  = YRELPOSITION_ID;
			break;
		case 10:
			name  = WIDTHREL_ID;
			break;
		case 11:
			name  = HEIGHTREL_ID;
			break;
		case 12:
			name  = FONTDELTA_ID;
			break;
		default:
			name  = super.getAttributeNameAt(index);
		}
		return(name);
	}

	public Object getAttributeValueAt(int index)
	{
		Object	value;

		switch (index) {
		case 0:
			value = getId();
			break;
		case 1:
			if (m_parentClass == null) {
				value = null;
			} else {
				value = m_parentClass.getId();
			}
			break;
		case 2:
			value = m_label;
			break;
		case 3:
			value = title;
			break;
		case 4:
			value = description;
			break;
		case 5:
			value = getObjectColor();
			break;
		case 6:
			value = getLabelColor();
			break;
		case 7:
			value = getObjectColorWhenOpen();
			break;
		case 8:
			value = new Double(xRelLocal());
			break;
		case 9:
			value = new Double(yRelLocal());
			break;
		case 10:
			value = new Double(widthRelLocal());
			break;
		case 11:
			value = new Double(heightRelLocal());
			break;
		case 12:
			value = new Integer(m_fontDelta);
			break;
		default:
			value = super.getAttributeValueAt(index);
		}
		return(value);
	}

	public void setAttributeValueAt(int index, Object value)
	{
		switch (index) {
		case 0:
			setId((String) value);
			break;
		case 1:
			if (value != null) {
				String newId = (String) value;
				if (m_parentClass == null || !m_parentClass.getId().equals(newId)) {
					Enumeration	en;
					EntityClass	ec;

					for (en = dg.enumEntityClasses(); en.hasMoreElements(); ) {
						ec = (EntityClass) en.nextElement();
						if (ec.getId().equals(newId)) {
							setParentClass(ec);
							break;
			}	}	}	}
			break;
		case 2:
			m_label = (String) value;
			break;
		case 3:
			title          = (String) value;
			break;
		case 4:
			description    = (String) value;
			break;
		case 5:
			setObjectColor((Color) value);
			break;
		case 6:
			setLabelColor((Color) value);
			break;
		case 7:
			setObjectColorWhenOpen((Color) value);
			break;
		case 8:
			setxRelLocal(((Double) value).doubleValue());
			break;
		case 9:
			setyRelLocal(((Double) value).doubleValue());
			break;
		case 10:
			setwidthRelLocal(((Double) value).doubleValue());
			break;
		case 11:
			setheightRelLocal(((Double) value).doubleValue());
			break;
		case 12:
			m_fontDelta    = ((Integer) value).intValue();
			break;
		default:
			super.setAttributeValueAt(index, value);
	}	}


	// Need to know the type in cases where value might be null
	// For example with some colors

	public int getAttributeTypeAt(int index)
	{
		int		ret;
		
		switch (index) {
		case 0:
		case 2:
		case 3:
			ret = Attribute.STRING_TYPE;
			break;
		case 1:
			ret = Attribute.ENTITY_CLASS_TYPE;
			break;
		case 4:
			ret = Attribute.TEXT_TYPE;
			break;
		case 5:
		case 6:
		case 7:
			ret = Attribute.COLOR_OR_NULL_TYPE;
			break;
		case 8:
		case 9:
		case 10:
		case 11:
			ret = Attribute.DOUBLE_TYPE;
			break;
		case 12:
			ret = Attribute.INT_TYPE;
			break;
		default:
			ret = super.getAttributeTypeAt(index);
		}
		return(ret);
	}

	public boolean containsDiagramPoint(int x, int y)
	{
		if (x < getDiagramX()) {
			return(false);
		}
		if (x > getDiagramX() + getWidth()) {
			return(false);
		}
		if (y < getDiagramY()) {
			return(false);
		}
		if (y > getDiagramY() + getHeight()) {
			return(false);
		}
		return(true);
	}

	// Icon interface (used to paint legends)

	public int getIconWidth()
	{
		return(getWidth());
	}

	public int getIconHeight()
	{
		return(getHeight());
	}

	public void paintIcon(Component c, Graphics g, int x, int y)
	{
		g.translate(x, y);
		paintComponent(g);
	}

	public void dump(int indent)
	{
		Enumeration		en;
		EntityInstance	e;
		int				i;

		for (i = 0; i < indent; ++i) {
			System.out.print(" ");
		}
		System.out.println(this);

		// Handle children

		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement();
			e.dump(indent+2);
		}
	}

	public int prepostorder(int value)
	{
		Enumeration		en;
		EntityInstance	e;

		m_preorder = value;

		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement();
			value = e.prepostorder(value+1);
		}
		m_postorder = ++value;
		return(value);
	}

	public void clearAllMarks(int preserve_entity_marks, int preserve_relation_marks) 
	{
		Enumeration			en;
		EntityInstance		e;
		RelationInstance	ri;

		andMark(preserve_entity_marks);
		
//		System.out.println("EntityInstance.clearAllMarks(" + preserve_entity_marks + ", " + preserve_relation_marks + ") " + m_label + " srcLifted empty" );
		m_srcLiftedList.removeAllElements();
		m_dstLiftedList.removeAllElements();
		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement();
			e.clearAllMarks(preserve_entity_marks, preserve_relation_marks);
		}

		// Doing all the source points clears all edges since all edges have a source

		for (en = m_srcRelList.elements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();
			ri.clearMark(preserve_relation_marks);
	}	}

	/* It is cheaper to visit each node outside of the draw root and not above it
	 * and see once if it has anything into the drawing, than to examine the edges
	 * inside the diagram to see where they go
	 *
	 * Adds to the vector set all those top most nodes outside of the draw root but
	 * not above it which have edges into the drawing
	 */

	public void addClients(Vector set, EntityInstance drawRoot)
	{
		Enumeration			en;
		RelationInstance	ri;
		EntityInstance		e;

		if (this != drawRoot) {
			// Don't descend beneath the drawroot
			if (!hasDescendant(drawRoot)) {
				for (en = m_srcRelList.elements(); en.hasMoreElements(); ) {
					ri  = (RelationInstance) en.nextElement();
					e   = ri.getDst();
					if (drawRoot.hasDescendant(e)) {
						orMark(CLIENT_MARK);
						set.addElement(this);
						return;
			}	}	}
			for (en = getChildren(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				e.addClients(set, drawRoot);
	}	}	}

	/* It is cheaper to visit each node outside of the draw root and not above it
	 * and see once if it has anything into the drawing, than to examine the edges
	 * inside the diagram to see where they go
	 *
	 * Adds to the vector set all those top most nodes outside of the draw root but
	 * not above it which have edges from the drawing
	 */

	public void addSuppliers(Vector set, EntityInstance drawRoot)
	{
		Enumeration			en;
		RelationInstance	ri;
		EntityInstance		e;

		if (this != drawRoot) {
			// Don't descend beneath the drawroot
			if (!hasDescendant(drawRoot)) {
				for (en = m_dstRelList.elements(); en.hasMoreElements(); ) {
					ri  = (RelationInstance) en.nextElement();
					e   = ri.getSrc();
					if (drawRoot.hasDescendant(e)) {
//						System.out.println("EntityInstance.addSuppliers " + this);
						orMark(SUPPLIER_MARK);
						set.addElement(this);
						return;
			}	}	}
			for (en = getChildren(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				e.addSuppliers(set, drawRoot);
	}	}	}

	protected void liftSrc(RelationInstance ri)
	{
		Enumeration			en;
		RelationInstance	other;

		if (!ri.isMarked(RelationInstance.PRESENTATION_MARKS)) {
			for (en = srcLiftedRelationElements(); en.hasMoreElements(); ) {
				other = (RelationInstance) en.nextElement();
				if (ri.parentClass == other.parentClass && ri.m_drawSrc == other.m_drawSrc && ri.m_drawDst == other.m_drawDst) {
					return;
		}	}	} 

		m_srcLiftedList.add(ri);	// Add this relation to the vector of relations that have draw source in this entity
	}

	protected void liftDst(RelationInstance ri)
	{
		Enumeration			en;
		RelationInstance	other;

		if (!ri.isMarked(RelationInstance.PRESENTATION_MARKS)) {
			for (en = dstLiftedRelationElements(); en.hasMoreElements(); ) {
				other = (RelationInstance) en.nextElement();
				if (ri.parentClass == other.parentClass && ri.m_drawSrc == other.m_drawSrc && ri.m_drawDst == other.m_drawDst) {
					return;
		}	}	} 
		m_dstLiftedList.add(ri);	// Add this relation to the vector of relations that have draw destination to this entity
	}

	// Mark any source relation whose destination is under 'under'

	public void orSrcRelationsDstUnder(EntityInstance under)
	{
		Enumeration			en;
		RelationInstance	ri;
		EntityInstance		dst;

		for (en = m_srcRelList.elements(); en.hasMoreElements(); ) {
			ri  = (RelationInstance) en.nextElement();
			dst = ri.getDst();
			if (under.hasDescendantOrSelf(dst)) {
				ri.orMark(RelationInstance.ADD_RESULTS_MARK);
	}	}	}

	public void orDstRelationsSrcUnder(EntityInstance under)
	{
		Enumeration			en;
		RelationInstance	ri;
		EntityInstance		src;

		for (en = m_dstRelList.elements(); en.hasMoreElements(); ) {
			ri  = (RelationInstance) en.nextElement();
			src = ri.getSrc();
			if (under.hasDescendantOrSelf(src)) {
				ri.orMark(RelationInstance.ADD_RESULTS_MARK);
	}	}	}

	public void orAllContainedSrcRelationsDstUnder(EntityInstance under)
	{
		Enumeration			en;
		EntityInstance		e;

		orSrcRelationsDstUnder(under);
		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement();
			e.orAllContainedSrcRelationsDstUnder(under);
	}	}

	public void addSrcRelations(int value, ResultBox resultBox)
	{
		Enumeration			en;
		RelationInstance	ri;

		for (en = m_srcRelList.elements(); en.hasMoreElements(); ) {
			ri  = (RelationInstance) en.nextElement();
			if (ri.isMarked(value)) {
				resultBox.addRelation(ri);
				ri.orMark(RelationInstance.HIGHLIGHT_FLAG_MARK);
				ri.nandMark(value);
	}	}	}

	public void addDstRelations(int value, ResultBox resultBox)
	{
		Enumeration			en;
		RelationInstance	ri;

		for (en = m_dstRelList.elements(); en.hasMoreElements(); ) {
			ri  = (RelationInstance) en.nextElement();
			if (ri.isMarked(value)) {
				resultBox.addRelation(ri);
				ri.orMark(RelationInstance.HIGHLIGHT_FLAG_MARK);
				ri.nandMark(value);
	}	}	}

	public void addSrcDstRelations(int value, ResultBox resultBox)
	{
		Enumeration			en;
		EntityInstance		e;

		addSrcRelations(value, resultBox);
		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement();
			e.addSrcDstRelations(value, resultBox);
	}	}

	/* Compute all the diagram edges before lifting them so that we can later when lifting them detect duplicate draw edges
	   and eliminate them.  That is we will set none null values for ri.m_drawSrc/ri.m_drawDst but not actually add them into
	   the entity vectors yet.
	 */

	public void computeAllDiagramEdges(EntityInstance drawRoot, Vector clients, Vector suppliers, EntityInstance closedEntity)
	{
		Enumeration			en;
		RelationInstance	ri;
		EntityInstance		src, dst, e;

//		System.out.println("EntityInstance.computeAllDiagramEdges for " + this);

		// For every edge that comes to us (something in or beneath entities in the diagram)
		for (en = m_dstRelList.elements(); en.hasMoreElements(); ) {
			ri  = (RelationInstance) en.nextElement();
			src = ri.m_src;
//			System.out.println(" -> " + ri);

			if (clients != null || suppliers != null) {
				// N.B. Clients may actually be classified as suppliers if one and the same
				if (!drawRoot.isPathBetween(src)) {
					ri.m_drawSrc = src.clientServer();
			}	}
			// The destination of ri comes to our entity

			if (closedEntity != null) {
				if (!closedEntity.hasDescendant(src)) {
					// Ignore internal loops beneath a closed entity
					ri.m_drawDst = closedEntity;
				}
			} else {
				// The draw destination of ri is its actual destination
				ri.m_drawDst = this;
			}
		}

		for (en = m_srcRelList.elements(); en.hasMoreElements(); ) {
			ri  = (RelationInstance) en.nextElement();
//			System.out.println(" <- " + ri);
			dst = ri.m_dst;

			if (suppliers != null && !drawRoot.isPathBetween(dst)) {
				ri.m_drawDst = dst.clientServer();
			}

			// The source of ri comes from our entity

			if (closedEntity != null) {
				if (!closedEntity.hasDescendant(dst)) {
					// Ignore internal loops beneath a close entity
					ri.m_drawSrc = closedEntity;
				}
			} else {
				// The source destination of ri is its actual destination
				ri.m_drawSrc = this;
			}
//			System.out.println(" <= " + ri);
		}

		if (closedEntity == null) {
			orMark(DIAGRAM_MARK);
			if (!isOpen()) {
				closedEntity = this;
		}	}

		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement();
			e.computeAllDiagramEdges(drawRoot, clients, suppliers, closedEntity);
	}	}
	
	public void liftAllDiagramEdges(EntityInstance drawRoot)
	{
		Enumeration			en;
		RelationInstance	ri;
		EntityInstance		e, src, dst;

		for (en = m_dstRelList.elements(); en.hasMoreElements(); ) {
			ri  = (RelationInstance) en.nextElement();
//			System.out.println("==>" + ri);
			if (!ri.isMarked(LIFTED_MARK)) {
				ri.orMark(LIFTED_MARK);
				src = ri.m_drawSrc;
				if (src != null) {
					dst = ri.m_drawDst;
					if (dst != null) {
						src.liftSrc(ri);
						dst.liftDst(ri);
					}
		}	}	}
		
		for (en = m_srcRelList.elements(); en.hasMoreElements(); ) {
			ri  = (RelationInstance) en.nextElement();
//			System.out.println("<==" + ri);
			if (!ri.isMarked(LIFTED_MARK)) {
				ri.orMark(LIFTED_MARK);
				src = ri.m_drawSrc;
				if (src != null) {
					dst = ri.m_drawDst;
					if (dst != null) {
						src.liftSrc(ri);
						dst.liftDst(ri);
		}	}	}	}
		
		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement();
			e.liftAllDiagramEdges(drawRoot);
	}	}

	public void validateEdges()
	{
		Enumeration			en;
		RelationInstance	ri;
		EntityInstance		e;

		for (en = dstLiftedRelationElements(); en.hasMoreElements(); ) {
			ri  = (RelationInstance) en.nextElement();
			ri.validate();
		}

		for (en = srcLiftedRelationElements(); en.hasMoreElements(); ) {
			ri  = (RelationInstance) en.nextElement();
			ri.validate();
		}

		if (!isMarked(CLOSED_MARK)) {
			for (en = getChildren(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				e.validateEdges();
	}	}	}

	public boolean intersectsLayout(Rectangle rect)
	{
		// Return true if the intersection of passed layout and 
		// entities layout is non-null

		// It overlaps if any corner of layout is contained in
		// entity, or vice versa.

		Rectangle box = new Rectangle(getX(), getY(), getWidth(), getHeight());
		return box.intersects(rect);
	}

 	public boolean containsRectangle(Rectangle rect) 
	{
		int x = getDiagramX();
		int y = getDiagramY();

		return (x < rect.x && rect.x+rect.width < x+getWidth() && y < rect.y && rect.y+rect.height < y+getHeight()); 
	}

	public int getParentDiagramX()
	{
		DiagramCoordinates above = (DiagramCoordinates) getParent();

		if (above == null) {
			return(0);
		}
		return(above.getDiagramX());
	}

	public int getParentDiagramY()
	{
		DiagramCoordinates above = (DiagramCoordinates) getParent();

		if (above == null) {
			return(0);
		}
		return(above.getDiagramY());
	}

	// Diagram Coordinates
	
	public int getDiagramX()
	{
		return(m_diagramX);
	}

	public int getDiagramY()
	{
		return(m_diagramY);
	}

	/* Compute the average position on the x axis of all drawn edges in/out of me */

	public double avgX() 
	{
		Enumeration			en;
		RelationInstance	ri;
		EntityInstance		src, dst;
		int					edgeMode;
		double				f;
		double				x = 0.0;
		int					n = 0;

		edgeMode = dg.getEdgeMode();
		if (edgeMode == Do.DIRECT_EDGE_STATE) {
			// For every edge coming out of us consider the destination
			for (en = srcLiftedRelationElements(); en.hasMoreElements(); ) {
				ri  = (RelationInstance) en.nextElement();
				dst = ri.m_drawDst;
				x  += (dst.getDiagramX() + dst.getWidth()/2);
				++n;
			}
			// For every edge coming into us consider the source
			for (en = dstLiftedRelationElements(); en.hasMoreElements(); ) {
				ri  = (RelationInstance) en.nextElement();
				src = ri.m_drawSrc;
				x  += (src.getDiagramX() + src.getWidth()/2);
				++n;
			}
		} else {
			for (en = srcLiftedRelationElements(); en.hasMoreElements(); ) {
				ri  = (RelationInstance) en.nextElement();
				dst = ri.m_drawDst;
				f   = ri.getRelationClass().getIOfactor();
				x  += (dst.getDiagramX() + dst.getWidth()*f);
				++n;
			}
			for (en = dstLiftedRelationElements(); en.hasMoreElements(); ) {
				ri  = (RelationInstance) en.nextElement();
				src = ri.m_drawSrc;
				f   = ri.getRelationClass().getIOfactor();
				x  += (src.getDiagramX() + src.getWidth()*f);
				++n;
		}	}
		if (n > 0) {
			return x/n;
		}
		return -1;
	}
}
